from typing import BinaryIO


class Buffer:
    def __init__(self, stream: BinaryIO):
        self.stream = stream

    def read(self, n: int) -> bytes:
        return self.stream.read(n)

    def read_int(self, n: int, ms_byte_first: bool = False) -> int:
        return int.from_bytes(
            self.read(n),
            byteorder='big' if ms_byte_first else 'little',
            signed=True,
        )

    def read_int8(self) -> int:
        return self.read_int(1)

    def read_int16(self, ms_byte_first: bool = False) -> int:
        return self.read_int(2, ms_byte_first)

    def read_int32(self, ms_byte_first: bool = False) -> int:
        return self.read_int(4, ms_byte_first)

    def read_uint(self, n: int, ms_byte_first: bool = False) -> int:
        return int.from_bytes(
            self.read(n),
            byteorder='big' if ms_byte_first else 'little',
            signed=False,
        )

    def read_uint8(self) -> int:
        return self.read_uint(1)

    def read_uint16(self, ms_byte_first: bool = False) -> int:
        return self.read_uint(2, ms_byte_first)

    def read_uint32(self, ms_byte_first: bool = False) -> int:
        return self.read_uint(4, ms_byte_first)

    def read_bool(self) -> bool:
        return self.read(1) != b'\x00'

    def read_string(self) -> str:
        data = bytearray()
        while True:
            b = self.read(1)
            if b == b'\x00' or b == b'':
                break
            data.extend(b)
        return data.decode('utf-8')

    def write(self, s: bytes) -> int:
        return self.stream.write(s)

    def write_int(self, i: int, n: int, ms_byte_first: bool = False) -> int:
        return self.write(i.to_bytes(
            n,
            byteorder='big' if ms_byte_first else 'little',
            signed=True,
        ))

    def write_int8(self, i: int) -> int:
        return self.write_int(i, 1)

    def write_int16(self, i: int, ms_byte_first: bool = False) -> int:
        return self.write_int(i, 2, ms_byte_first)

    def write_int32(self, i: int, ms_byte_first: bool = False) -> int:
        return self.write_int(i, 4, ms_byte_first)

    def write_uint(self, i: int, n: int, ms_byte_first: bool = False) -> int:
        return self.write(i.to_bytes(
            n,
            byteorder='big' if ms_byte_first else 'little',
            signed=False,
        ))

    def write_uint8(self, i: int) -> int:
        return self.write_uint(i, 1)

    def write_uint16(self, i: int, ms_byte_first: bool = False) -> int:
        return self.write_uint(i, 2, ms_byte_first)

    def write_uint32(self, i: int, ms_byte_first: bool = False) -> int:
        return self.write_uint(i, 4, ms_byte_first)

    def write_bool(self, b: bool) -> int:
        return self.write(b'\x01' if b else b'\x00')

    def write_nulls(self, n: int) -> int:
        for _ in range(n):
            self.write(b'\x00')
        return n

    def write_string(self, s: str) -> int:
        return self.write(s.encode('utf-8')) + self.write_nulls(1)

    def skip(self, n: int):
        self.seek(self.tell() + n)

    def seek(self, offset: int):
        self.stream.seek(offset)

    def tell(self) -> int:
        return self.stream.tell()
